/*
 Wireless Gameboy/GBA Controller - Multi-RX
 Version: 1.2a
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 3/10/2018
 Last Modified: 22/04/2019
 
 */

#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#ifndef NULL
#define NULL ((void *)0)
#endif

#define LOW 0
#define HIGH 1
#define false 0
#define true 1

#define LED PD7

#define GB_MODE 1
#define GBA_MODE 2

// Adapters
uint8_t adapterConnected = 0;
#define USB_CONNECTED 1
#define NES_CONNECTED 2
#define SNES_CONNECTED 3
#define GC_CONNECTED 4
#define WII_CONNECTED 5
#define N64_CONNECTED 6

// nRF settings (loaded from EEPROM)
uint8_t nrfChannel = 0;
uint8_t nrfRx[5] = "wgbr"; // Last byte loaded from EEPROM

// EEPROM settings
#define nrfChannelLocation 1
#define nrfRxLocation 2
#define blankEEPROMByte 255

// NES/SNES config
#define NES_SNES_CLK PC5
#define NES_SNES_DATA PC4
#define NES_SNES_LATCH PC3

// How many bit shifting loop times
uint16_t loopTimes = 0x8000; // SNES default

// GC config
#define GC_DATA_IN PC5
#define GC_DATA_OUT PC4

// WiiMote config - I2C
#define TWI_PORT C
#define TWI_SCL_PIN 5
#define TWI_SDA_PIN 4
#define twi_port PORTC
#define twi_ddr DDRC
#define twi_scl_pin TWI_SCL_PIN
#define twi_sda_pin TWI_SDA_PIN

// WiiMote ID
const unsigned char classic_controller_id[6] = {0x00, 0x00, 0xA4, 0x20, 0x01, 0x01};

// WiiMote Calibration data
const unsigned char cal_data[32] = {
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00
};

extern void wm_init(unsigned char *, unsigned char *, void (*)(void));
extern void wiimote_query(void);

// N64 config
#define N64_DATA_IN PC5
#define N64_DATA_OUT PC4

extern void mirf_init(void);
extern void mirf_config(void);


// SPI transfer 1 byte and return the result
uint8_t spi_transfer(uint8_t data) {
	SPDR = data;
	
	// Wait for transmission complete
	while (!(SPSR & (1<<SPIF)));
	
	return SPDR;
}

// Write data using SPI
void spi_write_data(uint8_t * dataout, uint8_t len) {
	uint8_t i;
	for (i = 0; i < len; i++) {
		spi_transfer(dataout[i]);
	}
}

// Read data using SPI
void spi_read_data(uint8_t * datain, uint8_t len) {
	uint8_t i;
	for (i = 0; i < len; i++) {
		datain[i] = spi_transfer(0x00);
	}
}

// Initialise the SPI
void spi_init(void) {
	DDRB |= (1<<PB5); // SPI CLK
	DDRB &= ~(1<<PB4); // SPI DI
	PORTB |= (1<<PB4); // SPI DI
	DDRB |= (1<<PB3); // SPI DO
	DDRB |= (1<<PB2); // SPI SS (nRF CSN)
	
	// Enable SPI, Master, set clock rate fck/4
	SPCR = (1<<SPE)|(1<<MSTR);
}

// Setup
void setup(void) {
	_delay_ms(200); // Wait a little bit
	
	// Check which adapter is connected
	uint8_t pinCRead = PINC & 0x07;
	if (pinCRead == 0) {
		adapterConnected = USB_CONNECTED;
	}
	else if (pinCRead == 1) {
		adapterConnected = SNES_CONNECTED;
		loopTimes = 0x8000;
		
		DDRC |= (1<<NES_SNES_DATA);
		PORTC |= (1<<NES_SNES_DATA); // Set high for no key presses
	}
	else if (pinCRead == 2) {
		adapterConnected = GC_CONNECTED;
		
		DDRC |= (1<<GC_DATA_OUT);
		PORTC &= ~(1<<GC_DATA_OUT);
	}
	else if (pinCRead == 3) {
		adapterConnected = NES_CONNECTED;
		loopTimes = 0x80;
		
		DDRC |= (1<<NES_SNES_DATA);
		PORTC |= (1<<NES_SNES_DATA); // Set high for no key presses
	}
	else if (pinCRead == 4) {
		adapterConnected = WII_CONNECTED;
	}
	else if (pinCRead == 5) {
		adapterConnected = N64_CONNECTED;
		
		DDRC |= (1<<N64_DATA_OUT);
		PORTC &= ~(1<<N64_DATA_OUT);
	}
	
	// Turn on pull-up resistors on other ports to save power
	PORTD |= (1<<PD0) | (1<<PD1) | (1<<PD5) | (1<<PD6);
	
	// Blink LED
	DDRD |= (1<<LED);
	PORTD |= (1<<LED);
	_delay_ms(50);
	PORTD &= ~(1<<LED);
	
	// Load nRF channel and address from EEPROM
	nrfChannel = eeprom_read_byte((uint8_t*) nrfChannelLocation);
	nrfRx[4] = eeprom_read_byte((uint8_t*) nrfRxLocation);
	if (nrfChannel == blankEEPROMByte) {
		nrfChannel = 1;
		nrfRx[4] = 1;
	}
	
	spi_init(); // Initialise SPI
	sei(); // Turn on interrupts
	
	mirf_init();
	_delay_ms(50);
	mirf_config();
}


// ******* Begin V-USB required code ********
const PROGMEM char usbHidReportDescriptor[42] = {
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x05,                    // USAGE (Game Pad)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x09, 0x01,                    //   USAGE (Pointer)
    0xa1, 0x00,                    //   COLLECTION (Physical)
    0x09, 0x30,                    //     USAGE (X)
    0x09, 0x31,                    //     USAGE (Y)
    0x15, 0x81,                    //     LOGICAL_MINIMUM (-127)
    0x25, 0x7f,                    //     LOGICAL_MAXIMUM (127)
    0x75, 0x08,                    //     REPORT_SIZE (8)
    0x95, 0x02,                    //     REPORT_COUNT (2)
    0x81, 0x02,                    //     INPUT (Data,Var,Abs)
    0xc0,                          //   END_COLLECTION
	 
    0x05, 0x09,                    // USAGE_PAGE (Button)
    0x19, 0x01,                    //   USAGE_MINIMUM (Button 1)
    0x29, 0x10,                    //   USAGE_MAXIMUM (Button 16)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x10,                    //   REPORT_COUNT (12)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0xc0                           // END_COLLECTION
};


usbMsgLen_t usbFunctionSetup(uchar data[8]) {
  usbRequest_t *req = (void *)data;

  return 0; // Nothing implemented
}

void usbEventResetReady(void) {
	//calibrateOscillator();
}

void hadUsbReset(void) {
  //calibrateOscillator();
  //eeprom_write_byte(0, OSCCAL);   /* store the calibrated value in EEPROM */
}

void startUSB(void) {
	usbDeviceDisconnect();
	_delay_ms(300); // 300 ms disconnect
	usbDeviceConnect();
	usbInit();
}