/*
 Wireless Gameboy/GBA Controller - Multi-RX
 Version: 1.2a
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 3/10/2018
 Last Modified: 22/04/2019
 
 */

#include <avr/io.h>
#include <avr/interrupt.h>

extern uint8_t modeFound;
extern uint16_t packetTimeout;
extern uint8_t kpData;
extern uint8_t kpDataHigh;
extern uint16_t volatile kpOutput;
extern uint8_t firstPacket;
extern uint8_t kpOutputMapYB;
extern uint8_t kpOutputMapYA;
extern uint8_t kpOutputMapRtoA;
extern uint8_t kpOutputMapLtoA;
extern uint8_t kpOutputUseJoystick;
extern uint8_t volatile timedOut;

// Gamecube
extern uint8_t volatile gcInputBuffer[10];
extern uint8_t kpOutputDpad;
extern uint8_t kpOutputZButton;
extern uint8_t kpOutputLRDownAndClick;

extern uint8_t kpLHeld;
extern uint8_t kpLHeldCounter;
extern uint8_t kpRHeld;
extern uint8_t kpRHeldCounter;

extern uint8_t gcReply1[3];
extern uint8_t gcReply2[10];
extern uint8_t gcControllerData[8];

// WiiMote Extension button presses
extern volatile int8_t jx, jy, rx, ry;
extern volatile uint8_t tl, tr;
extern volatile uint8_t left, right, up, down, a, b, x, y, 
						select, start, home, l, r, zl, zr;

// N64
extern uint8_t volatile n64ControllerData[4];

void nes_adapter(void) {
	// Check for key presses
	uint16_t tempKpOutput = 0xFFFF;
	if (kpData & 0x01) { // A
		tempKpOutput &= ~(1<<7);
	}
	if (kpData & 0x02) { // B
		tempKpOutput &= ~(1<<6);
	}
	if (kpData & 0x04) { // Right
		tempKpOutput &= ~(1<<0);
	}
	if (kpData & 0x08) { // Left
		tempKpOutput &= ~(1<<1);
	}
	if (kpData & 0x10) { // Up
		tempKpOutput &= ~(1<<3);
	}
	if (kpData & 0x20) { // Down
		tempKpOutput &= ~(1<<2);
	}
	if (kpData & 0x40) { // Select
		tempKpOutput &= ~(1<<5);
	}
	if (kpData & 0x80) { // Start
		tempKpOutput &= ~(1<<4);
	}
	
	// Assign new key presses
	kpOutput = tempKpOutput;
}

void snes_adapter(void) {
	// The first packet received triggers special modes
	if (firstPacket == 1) {
		firstPacket = 0;
		
		// A held down, map the Gameboy A button to the SNES Y button and Gameboy B button to the SNES A button
		if (kpData & 0x01) { // A
			kpOutputMapYB = 1;
		}
		// L held down, map GBA L button to the SNES A button
		if (kpDataHigh & 0x01) {
			kpOutputMapLtoA = 1;
		}
		// R held down, map GBA R button to the SNES A button
		if (kpDataHigh & 0x02) {
			kpOutputMapRtoA = 1;
		}
	}

	// Check for key presses
	uint16_t tempKpOutput = 0xFFFF;
	if (kpData & 0x01) { // A
		if (kpOutputMapYB == 1) {
			tempKpOutput &= ~(1<<15);
		}
		else {
			tempKpOutput &= ~(1<<7);
		}
	}
	if (kpData & 0x02) { // B
		if (kpOutputMapYB == 1) {
			tempKpOutput &= ~(1<<14);
		}
		else {
			tempKpOutput &= ~(1<<15);
		}
	}
	if (kpData & 0x04) { // Right
		tempKpOutput &= ~(1<<8);
	}
	if (kpData & 0x08) { // Left
		tempKpOutput &= ~(1<<9);
	}
	if (kpData & 0x10) { // Up
		tempKpOutput &= ~(1<<11);
	}
	if (kpData & 0x20) { // Down
		tempKpOutput &= ~(1<<10);
	}
	if (kpData & 0x40) { // Select
		tempKpOutput &= ~(1<<13);
	}
	if (kpData & 0x80) { // Start
		tempKpOutput &= ~(1<<12);
	}
	if (kpDataHigh & 0x01) { // L (for GBA)
		if (kpOutputMapLtoA == 1) {
			tempKpOutput &= ~(1<<7);
		}
		else {
			tempKpOutput &= ~(1<<5);
		}
	}
	if (kpDataHigh & 0x02) { // R (for GBA)
		if (kpOutputMapRtoA == 1) {
			tempKpOutput &= ~(1<<7);
		}
		else {
			tempKpOutput &= ~(1<<4);
		}
	}

	// Assign new key presses
	kpOutput = tempKpOutput;
}

void gc_adapter(void) {
	// The first packet received triggers special modes
	if (firstPacket == 1) {
		firstPacket = 0;
		
		// Down held down, output the key presses to the D-Pad
		if (kpData & 0x20) {
			kpOutputDpad = 1;
		}
		
		// R held down, simulate the L/R button presses like in the GC controller, if they are held for more than 200ms
		// then also trigger the click switch too. Useful for Gamecube games that might use this.
		if (kpDataHigh & 0x02) { 
			kpOutputLRDownAndClick = 1;
		}
		
		// Select held down, output the select key to the Z button
		if (kpData & 0x40) {
			kpOutputZButton = 1;
		}
		
		// A held down, B button now outputs Y and A button outputs B
		if (kpData & 0x01) { 
			kpOutputMapYB = 1;
		}
		
		// L held down, map GBA L button to the A button
		if (kpDataHigh & 0x01) {
			kpOutputMapLtoA = 1;
		}
	}
	
	// Default values
	uint8_t tempGcControllerData[8] = {0x00, 0x80, 0x7D, 0x78, 0x80, 0x7A, 0x1A, 0x1B};
	kpLHeld = 0;
	kpRHeld = 0;
	
	// Check for key presses
	if (kpData & 0x01) { // A
		if (kpOutputMapYB == 1) { // Map B button to A
			tempGcControllerData[0] |= 0x02;
		}
		else {
			tempGcControllerData[0] |= 0x01;
		}
	}
	if (kpData & 0x02) { // B
		if (kpOutputMapYB == 1) { // Map Y button to B
			tempGcControllerData[0] |= 0x08;
		}
		else {
			tempGcControllerData[0] |= 0x02;
		}
	}
	if (kpData & 0x04) { // Right
		if (kpOutputDpad == 1) {
			tempGcControllerData[1] |= 0x02; // D-Pad
		}
		else {
			tempGcControllerData[2] = 0xFF;
		}
	}
	if (kpData & 0x08) { // Left
		if (kpOutputDpad == 1) {
			tempGcControllerData[1] |= 0x01; // D-Pad
		}
		else {
			tempGcControllerData[2] = 0x00;
		}
	}
	if (kpData & 0x10) { // Up
		if (kpOutputDpad == 1) {
			tempGcControllerData[1] |= 0x08; // D-Pad
		}
		else {
			tempGcControllerData[3] = 0xFF;
		}
	}
	if (kpData & 0x20) { // Down
		if (kpOutputDpad == 1) {
			tempGcControllerData[1] |= 0x04; // D-Pad
		}
		else {
			tempGcControllerData[3] = 0x00;
		}
	}
	if (kpData & 0x40) { // Select
		if (kpOutputZButton == 0) {
			tempGcControllerData[0] |= 0x08; // Y
		}
		else {
			tempGcControllerData[1] |= 0x10; // Z
		}
	}
	if (kpData & 0x80) { // Start
		tempGcControllerData[0] |= 0x10;
	}
	if (kpDataHigh & 0x01) { // L (for GBA)
		if (kpOutputLRDownAndClick == 1) {
			kpLHeld = 1;
			kpLHeldCounter++;
			tempGcControllerData[6] = 0x78;
			
			if (kpLHeldCounter >= 250) {
				tempGcControllerData[1] |= 0x40;
				kpLHeldCounter = 250;
			}
		}
		else if (kpOutputMapLtoA == 1) {
			tempGcControllerData[0] |= 0x01;
		}
		else {
			tempGcControllerData[1] |= 0x40;
		}
	}
	if (kpDataHigh & 0x02) { // R (for GBA)
		if (kpOutputLRDownAndClick == 1) {
			kpRHeld = 1;
			kpRHeldCounter++;
			
			tempGcControllerData[7] = 0x78;
			if (kpRHeldCounter >= 250) {
				tempGcControllerData[1] |= 0x20;
				kpRHeldCounter = 250;
			}
		}
		else {
			tempGcControllerData[1] |= 0x20;
		}
	}
	
	// Reset L/R counter if not held down
	if (kpOutputLRDownAndClick == 1) {
		if (kpLHeld == 0) {
			kpLHeldCounter = 0;
		}
		if (kpRHeld == 0) {
			kpRHeldCounter = 0;
		}
	}
	
	// Assign new key presses
	for (uint8_t x = 0; x < 8; x++) {
		gcControllerData[x] = tempGcControllerData[x];
	}
}

void wii_adapter(void) {
	// The first packet received triggers special modes
	if (firstPacket == 1) {
		firstPacket = 0;
		
		// A held down, map the Gameboy A button to the SNES Y button and Gameboy B button to the SNES A button
		if (kpData & 0x01) {
			kpOutputMapYB = 1;
		}
		// B held down, map the Gameboy A button to the SNES Y button and Gameboy B button to the SNES A button
		if (kpData & 0x02) {
			kpOutputMapYA = 1;
		}
		// L held down, map GBA L button to the SNES A button
		if (kpDataHigh & 0x01) {
			kpOutputMapLtoA = 1;
		}
		// R held down, map GBA R button to the SNES A button
		if (kpDataHigh & 0x02) {
			kpOutputMapRtoA = 1;
		}
		// Down held down, use the joystick instead of the DPad
		if (kpData & 0x01) { // A
			kpOutputUseJoystick = 1;
		}
	}
	
	// Turn off interrupts before resetting key press data and assigning values
	cli();
	
	// Reset key presses
	left = 0, right = 0, up = 0, down = 0, a = 0, b = 0, x = 0, y = 0, 
	select = 0, start = 0, home = 0, l = 0, r = 0, zl = 0, zr = 0;
	
	// Check for key presses
	if (kpData & 0x01) { // A
		a = 1;
	}
	if (kpData & 0x02) { // B
		b = 1;
	}
	if (kpData & 0x04) { // Right
		right = 1;
	}
	if (kpData & 0x08) { // Left
		left = 1;
	}
	if (kpData & 0x10) { // Up
		up = 1;
	}
	if (kpData & 0x20) { // Down
		down = 1;
	}
	if (kpData & 0x40) { // Select
		select = 1;
	}
	if (kpData & 0x80) { // Start
		start = 1;
	}
	if (kpDataHigh & 0x01) { // L (for GBA)
		l = 1;
	}
	if (kpDataHigh & 0x02) { // R (for GBA)
		r = 1;
	}
	
	// Turn interrupts back on
	sei();
}

void n64_adapter(void) {
	// The first packet received triggers special modes
	if (firstPacket == 1) {
		firstPacket = 0;
		
		// Down held down, output the key presses to the D-Pad
		if (kpData & 0x20) {
			kpOutputDpad = 1;
		}
	}
	
	// Check for key presses
	uint8_t tempKpOutput[4] = {0x00, 0x00, 0x00, 0x00};
	if (kpData & 0x01) { // A
		tempKpOutput[0] |= 0x80;
	}
	if (kpData & 0x02) { // B
		tempKpOutput[0] |= 0x40;
	}
	if (kpData & 0x04) { // Right
		if (kpOutputDpad == 1) {
			tempKpOutput[0] |= 0x01;
		}
		else {
			tempKpOutput[2] = 80;
		}
	}
	if (kpData & 0x08) { // Left
		if (kpOutputDpad == 1) {
			tempKpOutput[0] |= 0x02;
		}
		else {
			tempKpOutput[2] = -80;
		}
	}
	if (kpData & 0x10) { // Up
		if (kpOutputDpad == 1) {
			tempKpOutput[0] |= 0x08;
		}
		else {
			tempKpOutput[3] = 80;
		}
	}
	if (kpData & 0x20) { // Down
		if (kpOutputDpad == 1) {
			tempKpOutput[0] |= 0x04;
		}
		else {
			tempKpOutput[3] = -80;
		}
	}
	if (kpData & 0x40) { // Select
		tempKpOutput[0] |= 0x20;
	}
	if (kpData & 0x80) { // Start
		tempKpOutput[0] |= 0x10;
	}
	if (kpDataHigh & 0x01) { // L (for GBA)
		tempKpOutput[1] |= 0x20;
	}
	if (kpDataHigh & 0x02) { // R (for GBA)
		tempKpOutput[1] |= 0x10;
	}
	
	// Assign new key presses
	n64ControllerData[0] = tempKpOutput[0];
	n64ControllerData[1] = tempKpOutput[1];
	n64ControllerData[2] = tempKpOutput[2];
	n64ControllerData[3] = tempKpOutput[3];
}

void usb_adapter(void) {
	usbPoll();

	if (usbInterruptIsReady()) {
		reportBuffer.data = 0x00; // Center pad
		
		// Check for key presses
		if (kpData & 0x01) { // A
			reportBuffer.A = 1;
		}
		if (kpData & 0x02) { // B
			reportBuffer.B = 1;
		}
		if (kpData & 0x04) { // Right
			reportBuffer.X_JOY = 127;
		}
		if (kpData & 0x08) { // Left
			reportBuffer.X_JOY = -127;
		}
		if (kpData & 0x10) { // Down
			reportBuffer.Y_JOY = -127;
		}
		if (kpData & 0x20) { // Up
			reportBuffer.Y_JOY = 127;
		}
		if (kpData & 0x40) { // Select
			reportBuffer.SELECT = 1;
		}
		if (kpData & 0x80) { // Start
			reportBuffer.START = 1;
		}
		if (kpDataHigh & 0x01) { // L (for GBA)
			reportBuffer.L = 1;
		}
		if (kpDataHigh & 0x02) { // R (for GBA)
			reportBuffer.R = 1;
		}
		
		// Called after every poll of the interrupt endpoint
		usbSetInterrupt(&reportBuffer, sizeof(reportBuffer));
	}
}